@@ -1,6 +1,6 @@
 /********************************************
 bi_funct.c
-copyright 2008-2014,2016, Thomas E. Dickey
+copyright 2008-2020,2021, Thomas E. Dickey
 copyright 1991-1995,1996, Michael D. Brennan
 
 This is a source file for mawk, an implementation of
@@ -11,7 +11,7 @@ the GNU General Public License, version 2, 1991.
 ********************************************/
 
 /*
- * $MawkId: bi_funct.c,v 1.110 2016/11/07 21:43:45 tom Exp $
+ * $MawkId: bi_funct.c,v 1.114 2021/05/28 23:38:12 tom Exp $
  */
 
 #include <mawk.h>
@@ -193,7 +193,6 @@ bi_index(CELL *sp)
 {
     size_t idx;
     size_t len;
-    const char *p;
 
     TRACE_FUNC("bi_index", sp);
 
@@ -202,6 +201,7 @@ bi_index(CELL *sp)
 	cast2_to_s(sp);
 
     if ((len = string(sp + 1)->len)) {
+	const char *p;
 	idx = (size_t) ((p = str_str(string(sp)->str,
 				     string(sp)->len,
 				     string(sp + 1)->str,
@@ -476,7 +476,6 @@ bi_strftime(CELL *sp)
     int n_args;
     int utc;
     STRING *sval = 0;		/* strftime(sval->str, timestamp, utc) */
-    char buff[128];
     size_t result;
 
     TRACE_FUNC("bi_strftime", sp);
@@ -517,7 +516,10 @@ bi_strftime(CELL *sp)
     else
 	ptm = localtime(&rawtime);
 
-    result = strftime(buff, sizeof(buff) / sizeof(buff[0]), format, ptm);
+    result = strftime(sprintf_buff,
+		      (size_t) (sprintf_limit - sprintf_buff),
+		      format,
+		      ptm);
     TRACE(("...bi_strftime (%s, \"%d.%d.%d %d.%d.%d %d\", %d) ->%s\n",
 	   format,
 	   ptm->tm_year,
@@ -528,12 +530,12 @@ bi_strftime(CELL *sp)
 	   ptm->tm_sec,
 	   ptm->tm_isdst,
 	   utc,
-	   buff));
+	   sprintf_buff));
 
     if (sval)
 	free_STRING(sval);
 
-    sp->ptr = (PTR) new_STRING1(buff, result);
+    sp->ptr = (PTR) new_STRING1(sprintf_buff, result);
 
     while (n_args > 1) {
 	n_args--;
@@ -948,7 +950,7 @@ bi_getline(CELL *sp)
     CELL tc;
     CELL *cp = 0;
     char *p = 0;
-    size_t len;
+    size_t len = 0;
     FIN *fin_p;
 
     TRACE_FUNC("bi_getline", sp);
@@ -1058,8 +1060,8 @@ bi_sub(CELL *sp)
     CELL *cp;			/* pointer to the replacement target */
     CELL tc;			/* build the new string here */
     CELL sc;			/* copy of the target CELL */
-    char *front, *middle, *back;	/* pieces */
-    size_t front_len, middle_len, back_len;
+    char *front, *middle;	/* pieces */
+    size_t middle_len;
 
     TRACE_FUNC("bi_sub", sp);
 
@@ -1083,9 +1085,9 @@ bi_sub(CELL *sp)
 		     0);
 
     if (middle != 0) {
-	front_len = (size_t) (middle - front);
-	back = middle + middle_len;
-	back_len = string(&sc)->len - front_len - middle_len;
+	size_t front_len = (size_t) (middle - front);
+	char *back = middle + middle_len;
+	size_t back_len = string(&sc)->len - front_len - middle_len;
 
 	if ((sp + 1)->type == C_REPLV) {
 	    STRING *sval = new_STRING0(middle_len);
@@ -1144,7 +1146,6 @@ gsub3(PTR re, CELL *repl, CELL *target)
 
     int skip0 = -1;
     size_t howmuch;
-    char *where;
 
     TRACE(("called gsub3\n"));
 
@@ -1161,11 +1162,11 @@ gsub3(PTR re, CELL *repl, CELL *target)
     output = new_STRING0(limit);
 
     for (j = 0; j <= (int) input->len; ++j) {
-	where = REmatch(input->str + j,
-			input->len - (size_t) j,
-			cast_to_re(re),
-			&howmuch,
-			(j != 0));
+	char *where = REmatch(input->str + j,
+			      input->len - (size_t) j,
+			      cast_to_re(re),
+			      &howmuch,
+			      (j != 0));
 	/*
 	 * REmatch returns a non-null pointer if it found a match.  But
 	 * that can be an empty string, e.g., for "*" or "?".  The length
